@@ -1,14 +1,17 @@
-from __future__ import absolute_import, division, print_function
+from __future__ import absolute_import
+from __future__ import division
+from __future__ import print_function
+
 import os
+import warnings
 
-import six
 import py
-import attr
+import six
 
-import _pytest
 import _pytest._code
-
+from _pytest.compat import getfslineno
 from _pytest.mark.structures import NodeKeywords
+from _pytest.outcomes import fail
 
 SEP = "/"
 
@@ -22,7 +25,7 @@ def _splitnode(nodeid):
         ''
         'testing/code'
         'testing/code/test_excinfo.py'
-        'testing/code/test_excinfo.py::TestFormattedExcinfo::()'
+        'testing/code/test_excinfo.py::TestFormattedExcinfo'
 
     Return values are lists e.g.
         []
@@ -30,11 +33,11 @@ def _splitnode(nodeid):
         ['testing', 'code', 'test_excinfo.py']
         ['testing', 'code', 'test_excinfo.py', 'TestFormattedExcinfo', '()']
     """
-    if nodeid == '':
+    if nodeid == "":
         # If there is no root node at all, return an empty list so the caller's logic can remain sane
         return []
     parts = nodeid.split(SEP)
-    # Replace single last element 'test_foo.py::Bar::()' with multiple elements 'test_foo.py', 'Bar', '()'
+    # Replace single last element 'test_foo.py::Bar' with multiple elements 'test_foo.py', 'Bar'
     parts[-1:] = parts[-1].split("::")
     return parts
 
@@ -42,36 +45,22 @@ def _splitnode(nodeid):
 def ischildnode(baseid, nodeid):
     """Return True if the nodeid is a child node of the baseid.
 
-    E.g. 'foo/bar::Baz::()' is a child of 'foo', 'foo/bar' and 'foo/bar::Baz', but not of 'foo/blorp'
+    E.g. 'foo/bar::Baz' is a child of 'foo', 'foo/bar' and 'foo/bar::Baz', but not of 'foo/blorp'
     """
     base_parts = _splitnode(baseid)
     node_parts = _splitnode(nodeid)
     if len(node_parts) < len(base_parts):
         return False
-    return node_parts[:len(base_parts)] == base_parts
-
-
-@attr.s
-class _CompatProperty(object):
-    name = attr.ib()
-
-    def __get__(self, obj, owner):
-        if obj is None:
-            return self
-
-        # TODO: reenable in the features branch
-        # warnings.warn(
-        #     "usage of {owner!r}.{name} is deprecated, please use pytest.{name} instead".format(
-        #         name=self.name, owner=type(owner).__name__),
-        #     PendingDeprecationWarning, stacklevel=2)
-        return getattr(__import__('pytest'), self.name)
+    return node_parts[: len(base_parts)] == base_parts
 
 
 class Node(object):
     """ base class for Collector and Item the test collection tree.
     Collector subclasses have children, Items are terminal nodes."""
 
-    def __init__(self, name, parent=None, config=None, session=None, fspath=None, nodeid=None):
+    def __init__(
+        self, name, parent=None, config=None, session=None, fspath=None, nodeid=None
+    ):
         #: a unique name within the scope of the parent node
         self.name = name
 
@@ -85,11 +74,14 @@ class Node(object):
         self.session = session or parent.session
 
         #: filesystem path where this node was collected from (can be None)
-        self.fspath = fspath or getattr(parent, 'fspath', None)
+        self.fspath = fspath or getattr(parent, "fspath", None)
 
         #: keywords/markers collected from all scopes
         self.keywords = NodeKeywords(self)
 
+        #: the marker objects belonging to this node
+        self.own_markers = []
+
         #: allow adding of extra keywords to use for matching
         self.extra_keyword_matches = set()
 
@@ -97,49 +89,51 @@ class Node(object):
         self._name2pseudofixturedef = {}
 
         if nodeid is not None:
+            assert "::()" not in nodeid
             self._nodeid = nodeid
         else:
-            assert parent is not None
-            self._nodeid = self.parent.nodeid + "::" + self.name
+            self._nodeid = self.parent.nodeid
+            if self.name != "()":
+                self._nodeid += "::" + self.name
 
     @property
     def ihook(self):
         """ fspath sensitive hook proxy used to call pytest hooks"""
         return self.session.gethookproxy(self.fspath)
 
-    Module = _CompatProperty("Module")
-    Class = _CompatProperty("Class")
-    Instance = _CompatProperty("Instance")
-    Function = _CompatProperty("Function")
-    File = _CompatProperty("File")
-    Item = _CompatProperty("Item")
-
-    def _getcustomclass(self, name):
-        maybe_compatprop = getattr(type(self), name)
-        if isinstance(maybe_compatprop, _CompatProperty):
-            return getattr(__import__('pytest'), name)
-        else:
-            cls = getattr(self, name)
-            # TODO: reenable in the features branch
-            # warnings.warn("use of node.%s is deprecated, "
-            #    "use pytest_pycollect_makeitem(...) to create custom "
-            #    "collection nodes" % name, category=DeprecationWarning)
-        return cls
-
     def __repr__(self):
-        return "<%s %r>" % (self.__class__.__name__,
-                            getattr(self, 'name', None))
+        return "<%s %s>" % (self.__class__.__name__, getattr(self, "name", None))
 
-    def warn(self, code, message):
-        """ generate a warning with the given code and message for this
-        item. """
-        assert isinstance(code, str)
-        fslocation = getattr(self, "location", None)
-        if fslocation is None:
-            fslocation = getattr(self, "fspath", None)
-        self.ihook.pytest_logwarning.call_historic(kwargs=dict(
-            code=code, message=message,
-            nodeid=self.nodeid, fslocation=fslocation))
+    def warn(self, warning):
+        """Issue a warning for this item.
+
+        Warnings will be displayed after the test session, unless explicitly suppressed
+
+        :param Warning warning: the warning instance to issue. Must be a subclass of PytestWarning.
+
+        :raise ValueError: if ``warning`` instance is not a subclass of PytestWarning.
+
+        Example usage::
+
+        .. code-block:: python
+
+            node.warn(PytestWarning("some message"))
+        """
+        from _pytest.warning_types import PytestWarning
+
+        if not isinstance(warning, PytestWarning):
+            raise ValueError(
+                "warning must be an instance of PytestWarning or subclass, got {!r}".format(
+                    warning
+                )
+            )
+        path, lineno = get_fslocation_from_item(self)
+        warnings.warn_explicit(
+            warning,
+            category=None,
+            filename=str(path),
+            lineno=lineno + 1 if lineno is not None else None,
+        )
 
     # methods for ordering nodes
     @property
@@ -167,26 +161,54 @@ class Node(object):
         chain.reverse()
         return chain
 
-    def add_marker(self, marker):
-        """ dynamically add a marker object to the node.
+    def add_marker(self, marker, append=True):
+        """dynamically add a marker object to the node.
 
-        ``marker`` can be a string or pytest.mark.* instance.
+        :type marker: ``str`` or ``pytest.mark.*``  object
+        :param marker:
+            ``append=True`` whether to append the marker,
+            if ``False`` insert at position ``0``.
         """
         from _pytest.mark import MarkDecorator, MARK_GEN
+
         if isinstance(marker, six.string_types):
             marker = getattr(MARK_GEN, marker)
         elif not isinstance(marker, MarkDecorator):
             raise ValueError("is not a string or pytest.mark.* Marker")
         self.keywords[marker.name] = marker
+        if append:
+            self.own_markers.append(marker.mark)
+        else:
+            self.own_markers.insert(0, marker.mark)
 
-    def get_marker(self, name):
-        """ get a marker object from this node or None if
-        the node doesn't have a marker with that name. """
-        val = self.keywords.get(name, None)
-        if val is not None:
-            from _pytest.mark import MarkInfo, MarkDecorator
-            if isinstance(val, (MarkDecorator, MarkInfo)):
-                return val
+    def iter_markers(self, name=None):
+        """
+        :param name: if given, filter the results by the name attribute
+
+        iterate over all markers of the node
+        """
+        return (x[1] for x in self.iter_markers_with_node(name=name))
+
+    def iter_markers_with_node(self, name=None):
+        """
+        :param name: if given, filter the results by the name attribute
+
+        iterate over all markers of the node
+        returns sequence of tuples (node, mark)
+        """
+        for node in reversed(self.listchain()):
+            for mark in node.own_markers:
+                if name is None or getattr(mark, "name", None) == name:
+                    yield node, mark
+
+    def get_closest_marker(self, name, default=None):
+        """return the first marker matching the name, from closest (for example function) to farther level (for example
+        module level).
+
+        :param default: fallback return value of no marker was found
+        :param name: name to filter by
+        """
+        return next(self.iter_markers(name=name), default)
 
     def listextrakeywords(self):
         """ Return a set of all extra keywords in self and any parents."""
@@ -218,6 +240,9 @@ class Node(object):
         pass
 
     def _repr_failure_py(self, excinfo, style=None):
+        if excinfo.errisinstance(fail.Exception):
+            if not excinfo.value.pytrace:
+                return six.text_type(excinfo.value)
         fm = self.session._fixturemanager
         if excinfo.errisinstance(fm.FixtureLookupError):
             return excinfo.value.formatrepr()
@@ -239,19 +264,47 @@ class Node(object):
             else:
                 style = "long"
 
+        if self.config.option.verbose > 1:
+            truncate_locals = False
+        else:
+            truncate_locals = True
+
         try:
             os.getcwd()
             abspath = False
         except OSError:
             abspath = True
 
-        return excinfo.getrepr(funcargs=True, abspath=abspath,
-                               showlocals=self.config.option.showlocals,
-                               style=style, tbfilter=tbfilter)
+        return excinfo.getrepr(
+            funcargs=True,
+            abspath=abspath,
+            showlocals=self.config.option.showlocals,
+            style=style,
+            tbfilter=tbfilter,
+            truncate_locals=truncate_locals,
+        )
 
     repr_failure = _repr_failure_py
 
 
+def get_fslocation_from_item(item):
+    """Tries to extract the actual location from an item, depending on available attributes:
+
+    * "fslocation": a pair (path, lineno)
+    * "obj": a Python object that the item wraps.
+    * "fspath": just a path
+
+    :rtype: a tuple of (str|LocalPath, int) with filename and line number.
+    """
+    result = getattr(item, "location", None)
+    if result is not None:
+        return result[:2]
+    obj = getattr(item, "obj", None)
+    if obj is not None:
+        return getfslineno(obj)
+    return getattr(item, "fspath", "unknown location"), -1
+
+
 class Collector(Node):
     """ Collector instances create children through collect()
         and thus iteratively build a tree.
@@ -274,7 +327,7 @@ class Collector(Node):
         return self._repr_failure_py(excinfo, style="short")
 
     def _prunetraceback(self, excinfo):
-        if hasattr(self, 'fspath'):
+        if hasattr(self, "fspath"):
             traceback = excinfo.traceback
             ntraceback = traceback.cut(path=self.fspath)
             if ntraceback == traceback:
@@ -285,7 +338,7 @@ class Collector(Node):
 def _check_initialpaths_for_relpath(session, fspath):
     for initial_path in session._initialpaths:
         if fspath.common(initial_path) == initial_path:
-            return fspath.relto(initial_path.dirname)
+            return fspath.relto(initial_path)
 
 
 class FSCollector(Collector):
@@ -306,10 +359,12 @@ class FSCollector(Collector):
 
             if not nodeid:
                 nodeid = _check_initialpaths_for_relpath(session, fspath)
-            if os.sep != SEP:
+            if nodeid and os.sep != SEP:
                 nodeid = nodeid.replace(os.sep, SEP)
 
-        super(FSCollector, self).__init__(name, parent, config, session, nodeid=nodeid, fspath=fspath)
+        super(FSCollector, self).__init__(
+            name, parent, config, session, nodeid=nodeid, fspath=fspath
+        )
 
 
 class File(FSCollector):
@@ -320,6 +375,7 @@ class Item(Node):
     """ a basic test invocation item. Note that for a single function
     there might be multiple test invocation items.
     """
+
     nextitem = None
 
     def __init__(self, name, parent=None, config=None, session=None, nodeid=None):
@@ -358,13 +414,7 @@ class Item(Node):
             return self._location
         except AttributeError:
             location = self.reportinfo()
-            # bestrelpath is a quite slow function
-            cache = self.config.__dict__.setdefault("_bestrelpathcache", {})
-            try:
-                fspath = cache[location[0]]
-            except KeyError:
-                fspath = self.session.fspath.bestrelpath(location[0])
-                cache[location[0]] = fspath
+            fspath = self.session._node_location_to_relpath(location[0])
             location = (fspath, location[1], str(location[2]))
             self._location = location
             return location
